    



class Int():
    a_type = "int"
    def __init__(self):
        pass

class Float():
    a_type = "float"
    def __init__(self):
        pass


class List():
    a_type = None
    def __init__(self):
        pass



class Tuple():
    length = None
    a_type = []
    def __init__(self):
        pass







"""
The privugger package
"""

from privugger.transformer import *
from privugger.measures import *
from privugger.inference import * 
from privugger.distributions import * 
from privugger.data_structures import * 





import pymc3 as pm
from scipy import stats as st
from abc import abstractmethod
"""
By specifying our own interface for distributions we could ideally hide which specific backend is used to model the distributions

"""

class Discrete():

    """
    An abstract class that represents the discrete distributions that privugger supports
    """
    @abstractmethod
    def pymc3_dist(self, name):
        return None

    @abstractmethod
    def get_params(self):
        return None
    
    @abstractmethod
    def scipy_dist(self, name):
        return None


__all__ = [

    "Bernoulli",
    "Categorical",
    "Binomial",
    "DiscreteUniform",
    "Geometric",
    "Constant"
]

#NOTE the convention is that num_elements -1 means that it is not set 
class Bernoulli(Discrete):
    """
    Class for the Bernoulli distribution
   
    Attributes 
    -----------
    
    name: String of the name of the random variable
    
    p: float value [0,1] giving the probability. Default: 0.5
    
    num_elements: int specifying number of RV's
    
    is_hyper_param: Boolean specifying if this RV is used as a hyper parameter. Default: False
    
    """

    def __init__(self,name, p=0.5, num_elements=-1, is_hyper_param=False):


        self.p = p
        self.name =name
        self.num_elements=num_elements
        self.is_hyper_param = is_hyper_param
    
    def pymc3_dist(self, name, hypers):
        p = self.p
        if(len(hypers) == 1):
                hyper_dist = hypers[0][0]
                hyper_name = hypers[0][1]
                p = hyper_dist.pymc3_dist(hyper_name, [])

        if(self.num_elements==-1):
            return pm.Bernoulli(name, p=p)
        else:
            return pm.Bernoulli(name, p=p, shape=self.num_elements)

    def get_params(self):
        return [self.p]
    
    def scipy_dist(self, name):
        dist = (lambda siz : st.bernoulli(p=self.p).rvs(siz)) if self.num_elements == -1 else (lambda siz: st.bernoulli(p=self.p).rvs((self.num_elements, siz)))
        return name,dist



class Categorical(Discrete):

    """
    Class for the Categorical distribution
    
    Attributes 
    -----------
   
    name: String of the name of the random variable
   
    p: Float list of probabilities
   
    num_elements: int specifying number of RV's
   
    is_hyper_param: Boolean specifying if this RV is used as a hyper parameter. Default: False
   
    """


    
    def __init__(self, name, p=None, num_elements=-1, is_hyper_param=False):

        if (p==None):
            raise TypeError("please specify p")
        else:
            self.p=p

        self.num_elements=num_elements
        self.name = name
        self.is_hyper_param = is_hyper_param

    def pymc3_dist(self, name, hypers):
        p = self.p
        if(len(hypers) == 1):
            hyper_dist = hypers[0][0]
            hyper_name = hypers[0][1]
            p = hyper_dist.pymc3_dist(hyper_name, [])
            
        if(self.num_elements==-1):
            return pm.Categorical(name, p=p)
        else:
            return pm.Categorical(name, p=p, shape=self.num_elements)

    def get_params(self):
        return [self.p]
    
    def scipy_dist(self, name):
        theta = self.p
        dist = (lambda siz : st.rv_discrete(values=(range(len(theta)), theta)).rvs(siz)) if self.num_elements == -1 else (lambda siz: st.rv_discrete(values=(range(len(theta)), theta)).rvs((self.num_elements, siz)))
        return name, dist


class Binomial(Discrete):

    """
    Class for the Binomial distribution
    
    Attributes 
    -----------
    
    name: String of the name of the random variable
    
    n: int specifying the number of trials. Default: 2
    
    p: float value [0,1] giving the probability. Default: 0.5
    
    num_elements: int specifying number of RV's
    
    is_hyper_param: Boolean specifying if this RV is used as a hyper parameter. Default: False
    
    """


    
    def __init__(self, name, n=2, p=0.5, num_elements=-1, is_hyper_param=False):

        self.n=n
        self.name = name
        self.p=p
        self.num_elements=num_elements
        self.is_hyper_param = is_hyper_param
    
    def pymc3_dist(self, name, hypers):
        n = self.n
        p = self.p

        if(len(hypers) == 1):
                hyper_dist = hypers[0][0]
                hyper_name = hypers[0][1]
                idx = hypers[0][2]
                if(idx == 0):
                    n = hyper_dist.pymc3_dist(hyper_name, [])
                else:
                    p = hyper_dist.pymc3_dist(hyper_name, [])
        elif(len(hypers) == 2):
                hyper_dist_1 = hypers[0][0]
                hyper_name_1 = hypers[0][1]
                hyper_dist_2 = hypers[1][0]
                hyper_name_2 = hypers[1][1]
                n = hyper_dist_1.pymc3_dist(hyper_name_1, [])
                p = hyper_dist_2.pymc3_dist(hyper_name_2, [])


        if(self.num_elements==-1):
            return pm.Binomial(name, n=n, p=p)
        else:
            return pm.Binomial(name, n=n, p=p, shape=self.num_elements)

    def get_params(self):
        return [self.n, self.p]
    
    def scipy_dist(self, name):
        dist = (lambda siz : st.binom(n=self.n, p=self.p).rvs(siz)) if self.num_elements == -1 else (lambda siz: st.binom(n=self.n, p=self.p).rvs((self.num_elements, siz)))
        return name, dist

class DiscreteUniform(Discrete):

    """
    Class for the Discrete Uniform distribution
    
    Attributes 
    -----------
    
    name: String of the name of the random variable
    
    lower: int value giving the lower bound of the values. Default: 0
    
    upper: ine value giving the upper bound of the values. Default: 1
    
    num_elements: int specifying number of RV's
    
    is_hyper_param: Boolean specifying if this RV is used as a hyper parameter. Default: False
    
    """

    def __init__(self, name, lower=0, upper=1, num_elements=-1, is_hyper_param=False):
  


        self.lower = lower
        self.upper = upper
        self.name = name
        self.num_elements=num_elements
        self.is_hyper_param = is_hyper_param

    def pymc3_dist(self, name, hypers):
        lower = self.lower
        upper = self.upper
        if(len(hypers) == 1):
                hyper_dist = hypers[0][0]
                hyper_name = hypers[0][1]
                idx = hypers[0][2]
                if(idx == 0):
                    lower = hyper_dist.pymc3_dist(hyper_name, [])
                else:
                    upper = hyper_dist.pymc3_dist(hyper_name, [])
        elif(len(hypers) == 2):
                hyper_dist_1 = hypers[0][0]
                hyper_name_1 = hypers[0][1]
                hyper_dist_2 = hypers[1][0]
                hyper_name_2 = hypers[1][1]
                lower = hyper_dist_1.pymc3_dist(hyper_name_1, [])
                upper = hyper_dist_2.pymc3_dist(hyper_name_2, [])


        if(self.num_elements==-1):
            return pm.DiscreteUniform(name, lower=lower, upper=upper)
        else:
            return pm.DiscreteUniform(name, lower=lower, upper=upper, shape=self.num_elements)

    def get_params(self):
        return [self.lower, self.upper]

    def scipy_dist(self, name):
        dist = (lambda siz : st.randint(lower=self.lower, upper=self.upper).rvs(siz)) if self.num_elements == -1 else (lambda siz: st.randint(lower=self.lower, upper=self.upper).rvs((self.num_elements, siz)))
        return name, dist

class Geometric(Discrete):

    """
    Class for the Geometric distribution
    
    Attributes 
    -----------
    
    name: String of the name of the random variable
    
    p: float value [0,1] giving the probability. Default: 0.5
    
    num_elements: int specifying number of RV's
    
    is_hyper_param: Boolean specifying if this RV is used as a hyper parameter. Default: False
    
    """


    
    def __init__(self, p=0.5, num_elements=-1, is_hyper_param=False):
   
        self.p=p
        self.num_elements=num_elements
        self.is_hyper_param = is_hyper_param

    def pymc3_dist(self, name, hypers):
        p = self.p
        if(len(hypers) == 1):
            hyper_dist = hypers[0][0]
            hyper_name = hypers[0][1]
            p = hyper_dist.pymc3_dist(hyper_name, [])

        if(self.num_elements==-1):
            return pm.Geometric(name, p=p)
        else:
            return pm.Geometric(name, p=p, shape=self.num_elements)
        
    def get_params(self):
        return [self.p]
    
    def scipy_dist(self, name):
        dist = (lambda siz : st.geom(self.p).rvs(siz)) if self.num_elements == -1 else (lambda siz: st.geom(self.p).rvs((self.num_elements, siz)))
        return name, dist


class Constant(Discrete):

    """
    Class for the Constant distribution
    
    Attributes 
    -----------
    
    name: String of the name of the random variable
    
    val: The constant value 
    
    num_elements: int specifying number of RV's
    
    is_hyper_param: Boolean specifying if this RV is used as a hyper parameter. Default: False
    
    """


    
    def __init__(self, name, val, num_elements=-1, is_hyper_param=False):

        self.val = val
        self.name = name
        self.num_elements = num_elements
        self.is_hyper_param = is_hyper_param

    def pymc3_dist(self, name, hypers):
        val = self.val
        if(len(hypers) == 1):
            hyper_dist = hypers[0][0]
            hyper_name = hypers[0][1]
            val = hyper_dist.pymc3_dist(hyper_name, [])
        if(self.num_elements==-1):
            return pm.ConstantDist(name, self.val)
        else:
            return pm.ConstantDist(name, self.val, shape=self.num_elements)

    def get_params(self):
        return [self.val]
    
    def scipy_dist(self, name):
        return lambda siz: np.array([self.val for _ in range(siz)])
















import theano
import theano.tensor as tt
import numpy as np
from typing import List


@theano.compile.ops.as_op(itypes=[tt.lscalar], otypes=[tt.lscalar])
def original_pwd_checker(password: int) ->int:
    PWD = 1024
    return np.int64(password == PWD)




import numpy as np
import pandas as pd
import portion as p
from scipy.special import kl_div, entr, rel_entr

    
""" 
    Parameters
    ------------
    P = Array of samples  array or list
    Q = Array of samples, array or list

    l_m = Partion size, integer
    dim = Dimension of sample, integer
    step = Step size, integer or float

"""
 
 
 

def discrete(P, Q):
    H = lambda P: entr(P)
    H_cross = lambda P, Q: rel_entr(P, Q)
    kl_divergence = lambda P, Q: kl_div(P, Q)

    return  pd.Series({'Entropy':sum(H(P)), 'Cross Entropy': sum(H_cross(P,Q)), 'KL divergence':sum(kl_divergence(P,Q))})

  
  
def continuous(P, Q, l_m, dim=1, step=0.01):
        
    m = np.size(P)
    n = np.size(Q)
        
    t_m = int(np.power((m/l_m),(1/dim)))

    dt = l_m*(t_m-1)

    def I_mid(X):
            
        i = l_m
        j = 2*l_m
            
        I = list(p.iterate(p.openclosed(X[i], X[j]), step=step))
            
        while  j == dt is False:
            i = 2*i
            j = 2*j
            I = I.append(list(p.iterate(p.openclosed(X[i], X[j]), step=step)))
        else:
            pass
        return I
        
    I_start = lambda X: list(p.iterate(p.openclosed(min(X), X[l_m]), step=step))

    I_end = lambda X: list(p.iterate(p.open(X[dt], max(X)), step=step))
        
    I = lambda X:[*I_start(X),*I_mid(X),*I_end(X)]

    P_n = I(np.sort(P, axis=None))
    Q_m = I(np.sort(Q, axis=None))

    D_kl = sum([P_n[i]*np.log(P_n[i]/Q_m[i]) for i in range(0, np.size(P_n))])

    return pd.Series({'KL Divergence': D_kl})
       




from privugger.data_structures.dataset import *
import re
import pymc3 as pm

class Program:

    def __init__(self, name, dataset, output_type, function):
        """
        A class representing the privacy preserving program to be analysed.

        Parameters
        ------------
        dataset: the dataset of type privugger.Dataset containg the values used in the program
        output_type: the output type specified as Int, Float, List(Int), List(Float)
        program: The program to be analysed, either string to location program, lambda method or def function
        """
        if isinstance(dataset, Dataset):
            self.dataset              = dataset
            self.output_type          = output_type
            self.name                 = name
            self.program              = function
            self.observation          = None
            self.execute_observations = lambda a,b: None
        else:
            raise ValueError("The dataset has to be of type privugger.Dataset")

        
    def add_observation(self, constraints, precision=0.01):
        """Adds observation based on a string specifying inequalities of the formed

        `a X var` or `a X var X b` where a,b are of type Int or a,b,c
        are of type Float, `var` is a string corresponding to one of
        the variables in the program or its output, and X \in
        {<,>,=<,>=,==}.
        
        Examples: 
          - 10 > output > 5
          - 52.5 < output
          - output == 42.5 

        Parameters
        ------------
        constraints : String 
            A string representing the constraints to be added

        precision : Float 
            Models the precision with which the observation must
            hold. A value of 0 requires the observation to hold with
            probability 1. The larger the value the lower the
            probability required for the condition to hold.

        """
        cons = "[-+]?([0-9]*\.[0-9]+|[0-9]+)*([>=<]*)([a-zA-Z\s]*)([>=<]*)[-+]?([0-9]*\.[0-9]+|[0-9]+)*"
        constraints = constraints.replace(" ", "")
        vals = re.search(cons, constraints)

        # val1 cons1 name cons2 val2
        # 10 > output >= 2
        val1 = vals.group(1)
        cons1 = vals.group(2)
        name = vals.group(3)
        cons2 = vals.group(4)
        val2 = vals.group(5)

        # DEBUGGING
        # print("val1: " + str(val1) + " cons1: " + str(cons1) + " name: " + str(name) + " cons2: " + str(cons2) + " val2: " + str(val2))

        # Well-formedness checks
        
        ## Check whether we have something of the type `a X var X`
        if val2 == None and cons2 != "":
            raise ValueError("Malformed observation: please review the observation to ensure that the syntax is correct.")

        ## TODO: Check that the type of the a, b in `a X var X b` match the type of var
        
        

        var_names  = self.dataset._collect_distribution_names()
        
        if name.strip() in var_names:
            name = name.strip()

        partial1 = lambda x: None
        partial2 = lambda x: None
        if name in var_names or "output" in name:
            if val1 != "" and cons1 != "":
                v1 = float(val1) if "." in val1 else int(val1)
                partial1 = self._unwrap_constrain(v1, cons1, precision)
                print(partial1)
                
            if val2 != "" and cons2 != "":
                v2 = float(val2) if "." in val2 else int(val2)
                partial2 = self._unwrap_constrain(v2, cons2, precision, i=1)
        
            def inner(prior, output):
                if name in var_names:
                    idx = var_names.index(name)
                    distribution = prior[idx]
                else:
                    distribution = output
                partial1(distribution)
                partial2(distribution)
            self.execute_observations = inner
            return None # to avoid having a return value
        else:
            raise ValueError("Observation was not known. Make sure that the name is part of the names in privugger.Datastructure")


    def _unwrap_constrain(self, value, cons, precision, i=0):
        if not i % 2:
            cons = cons.replace(">", "<")
        def inner(distribution):
            if cons == ">":
                pm.Normal(f"cons_{i}", distribution>value,  precision, observed=1)
            elif cons == ">=":
                pm.Normal(f"cons_{i}", distribution>=value, precision, observed=1)
            elif cons == "<":
                pm.Normal(f"cons_{i}", distribution<value,  precision, observed=1)
            elif cons == "<=":
                pm.Normal(f"cons_{i}", distribution<=value, precision, observed=1)
            elif cons == "==":
                pm.Normal(f"cons_{i}", distribution,        precision, observed=value)
            else:
                raise ValueError(f"The program does not support {cons} as a constrain")
        return inner






__all__ = [
    "Dataset"
]

class Dataset:

    def __init__(self, input_specs):

        """
        Dataset represents the specs about the input data 
        
        Parameters
        -------------
        input_spec: specification about the distributions of the input given as a list of distributions

        program_output: the output of the program as a standard type
        
        """
        self.input_specs = input_specs

    def _collect_distribution_names(self):
        """
        
        Convenience method for collecting names associated with the distributions 
        
        Returns
        -------------
        names: A list containing all the names of the random variables in the dataset
        """
        names = []
        for distribution in self.input_specs:
            names.append(distribution.name)
        return names
        
    
class Float:
    
    def __init__(self, dist= None,  name=None):
        """
        Float represents a floating point value drawn from a spcified distribution
    
        Parameters
        -----------
        dist: distribution 
        name: name associated with distrbution
        
        """
        self.dist=dist
        if(name is None):
            raise ValueError("name must be specified")
        else:
            self.name=name

   
class Int:

    def __init__(self, dist=None,  name=None):
        """
        Int represents an integer value drawn from a spcified distribution

        Parameters
        -----------
        dist: distribution 
        name: name associated with distrbution

        """
        self.dist = dist
        if(name is None):
            raise ValueError("name must be specified")
        else:
            self.name = name


            



from hypothesis import settings, given, Phase, HealthCheck, strategies as st
import numpy as np
import pymc3 as pm
from privugger.attacker.generators import IntGenerator, IntList, FloatGenerator, FloatList, DiscreteUniform, Uniform
from privugger.attacker.metrics import SimulationMetrics
from sklearn.feature_selection import mutual_info_regression
import typing
import inspect

"""
The data privacy debugger, PRIVUGER, is a privacy risk analysis tool.
"""

def simulate(function, *args, **kwargs) -> SimulationMetrics:
    """
    The main method used to simulate a method

    **Returns: SimulationMetrics **
    ----------
        - Returns a SimulationMetrics, a method helped to assist the data analyst

    **Parameters:**
    ----------
    Types: Types
        - The types that your method takes
        - Example: Tuple[int, float]
    *number_of_test: int*
        - Number of test to be executed
        - Default: 1
    *size: int*
        - Size of the database to simulate
        - Default: 4
    *samples: int*
        - Number of samples per. execution
        - Default: 1000
    *ranges: list[tuple[int,int]]*
        - A list of ranges that the distributions should mimic
        - Default: (0, 100)
    """
    # Check that all parameters have type annotation:
    if len(inspect.signature(function).parameters) != len(function.__annotations__)-1:
        raise TypeError("You need to specify all types in your method before analyzing")

    # The output trace:
    traces = []

    #Values for KWARGS
    max_examples = 1 if "max_examples" not in kwargs else kwargs["max_examples"]
    N = 2 if "N" not in kwargs else kwargs["N"]
    samples = 1000 if "num_samples" not in kwargs else kwargs["num_samples"]
    current_test = []
    ranges = [(-10000, 10000) for _ in range(len(inspect.signature(function).parameters))] if "ranges" not in kwargs else kwargs["ranges"]
    logging = True if "logging" not in kwargs else kwargs["logging"]

    #Logging
    if not logging:
        import logging
        logger = logging.getLogger("pymc3")
        logger.setLevel(logging.ERROR)
        logger.propagate = False

    @settings(max_examples=max_examples, deadline=None, phases=[Phase.generate], suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(st.data())
    def helper(data):

        def parse(argument, islist=False, istuple=False, parameter_pos=0, ranges=ranges):
            """
            PSEUDO CODE FOR PARSE:
            values <- Ø
            For each p <- parameter type in argument
                case p == list
                    x <- Ø
                    s <- Construct a secret to cover all possible values as p
                    x[0] <- s
                    for n=1, n<=N, n++
                        d <- parse(p)
                return x
                case p == tuple
                    x <- Ø
                    for t <- type in p
                        x += parse(t)
                case p == float
                    d <- float generator
                    return d
                case p == int
                    d <- int generator
                    return d
            """
            if argument == int:
                name = f"Alice-int_{parameter_pos}"
                if islist:
                    #Figure out how to fetch range:
                    alice_int = pm.distributions.DiscreteUniform(name, ranges[0][0], ranges[0][1])
                    dist, info = IntList(name=f"intList_{parameter_pos}", data=data, length=N, ranges=ranges[0])
                    return ((alice_int, dist,info), parameter_pos)
                else:
                    dist,info = IntGenerator(data=data, name=f"intDist_{parameter_pos}", ranges=ranges[0])
                    if not istuple:
                        return (([dist], dist,info), parameter_pos)
                    alice_int = pm.distributions.DiscreteUniform(name, ranges[0][0], ranges[0][1])
                    return ((alice_int, dist,info), parameter_pos)
            elif argument == float:
                if islist:
                    alice_float = pm.distributions.Uniform(f"Alice-float_{parameter_pos}", ranges[0][0], ranges[0][1])
                    dist, info = FloatGenerator(name=f"floatList_{parameter_pos}", data=data, shape=N, ranges=ranges[0])
                    return ((alice_float, dist,info), parameter_pos)
                else:
                    dist, info = FloatGenerator(name=f"floatDist_{parameter_pos}", data=data, shape=1, ranges=ranges[0])
                    if not istuple:
                        return (([dist], dist,info), parameter_pos)
                    alice_float = pm.distributions.Uniform(f"Alice-float_{parameter_pos}", ranges[0][0], ranges[0][1])
                    return ((alice_float, dist,info), parameter_pos)
            elif argument.__origin__ == list or argument.__origin__ == typing.List:
                # parameter is a list
                x = np.empty(N+1, dtype=object)
                alice_info, d, inf = [],[],[]
                for i, p in enumerate(argument.__args__):
                    (alice, dist, info), pos = parse(p, islist=True, istuple=istuple, parameter_pos=parameter_pos, ranges=ranges[i:])
                    parameter_pos = pos+1
                    alice_info = alice
                    d.append(dist)
                    inf.append(info)
                x[0] = alice
                inputs = [[] for _ in range(N)]
                for k in range(N):
                    for par in d:
                        if len(par) > 1:
                            for dist in par:
                                try:
                                    inputs[k].append(dist[k])
                                except:
                                    break
                        else:
                            inputs[k].append(par[k])
                    x[k+1] = inputs[k]
                return ((alice_info, x, info), parameter_pos)
            elif argument.__origin__ == tuple or argument.__origin__ == typing.Tuple:
                alice_info, d, i = [],[],[]
                for j, p in enumerate(argument.__args__):
                    (alice, dist, info),pos = parse(p, islist=islist, istuple=True, parameter_pos=parameter_pos, ranges=ranges[j:])
                    parameter_pos = pos+1
                    alice_info.append(alice)
                    d.append(dist)
                    i.append(info)
                return ((tuple(alice_info), tuple(d), tuple(i)), parameter_pos)

        #Progress bar
        current_test.append([])
        percentage = (len(current_test)/max_examples)
        width = 100
        fill = "█"*(int(width*percentage))
        tail = "-"*(int(width*(1-percentage)))
        print("\r["+fill+tail+"] " + str(len(current_test)) + "/" + str(max_examples), end="\r")

        #Parameters setup
        parameters = list(function.__annotations__.values())
        pos = 0
        alice_names = []
        outputs = []
        info = []

        #A single data type
        if len(parameters) == 2:
            with pm.Model() as model:
                p = parameters[0]
                (alice, dist, temp_info), pos = parse(p, parameter_pos=pos, ranges=ranges)
                for n in alice:
                    alice_names.append(n)
                outputs.append(dist)
                info.append(temp_info)
                pos+=1
                output = pm.Deterministic("Output", function(*outputs))
                trace = pm.sample(samples, cores=1)
                traces.append((trace, alice_names, info))      
        else:
            TEST_PER_PARAMETER = 10
            # generate for the first parameter and keep the other locked
            for p_pos, p in enumerate(parameters[:-1]):
                pos = 0
                with pm.Model() as model:
                    # p becomes the one to test on
                    alice_info_pairs = [[] for _ in range(TEST_PER_PARAMETER)]
                    info_pairs = []
                    output_pairs = []
                    # Generate 10 samples to test versus one specific distribution
                    for i in range(TEST_PER_PARAMETER):
                        (alice, dist, temp_info), pos = parse(p, parameter_pos=pos)
                        for n in alice:
                            alice_info_pairs[i].append(n)
                        output_pairs.append([dist])
                        info_pairs.append([temp_info])
                        pos+=1
                    for p2_pos, p_2 in enumerate(parameters[:-1]):
                        if p2_pos != p_pos:
                            (alice, dist, temp_info), pos = parse(p, parameter_pos=pos)
                            for i in range(TEST_PER_PARAMETER):
                                for n in alice:
                                    alice_info_pairs[i].append(n)
                                output_pairs[i].append(dist)
                                info_pairs[i].append(temp_info)
                            pos+=1
                    for i in range(TEST_PER_PARAMETER):
                        pm.Deterministic(f"Output_{i}", function(*output_pairs[i]))
                    trace = pm.sample(samples, cores=1, steps=pm.NUTS())
                    for i in range(TEST_PER_PARAMETER):
                        traces.append((trace, alice_info_pairs[i], info_pairs[i]))   
    helper()
    return SimulationMetrics(traces)

def simulate_decorator(*args, **kwargs):
    """
    ***A decorator used to probabilistically analyse the method***

    **Returns: List[float] **
    ----------
        - Returns a list containing mutual information based on each test

    **Parameters:**
    ----------
    Types: Types
        - The types that your method takes
        - Example: Tuple[int, float]
    *number_of_test: int*
        - Number of test to be executed
        - Default: 1
    *size: int*
        - Size of the database to simulate
        - Default: 4
    *samples: int*
        - Number of samples per. execution
        - Default: 1000
    *ranges: list[tuple[int,int]]*
        - A list of ranges that the distributions should mimic
        - Default: (0, 100)
    """
    def inner(func):
        trace = simulate(func, args, kwargs)
        return (lambda x=trace:x)
    return inner







__all__ =[
    "Int",
    "Float",
    "List",
    "Matrix"
]

class Output: pass

class Int(Output): pass
class Float(Output): pass
class List(Output):
    def __init__(self, output):
        self.output = output
class Matrix(Output):
    def __init__(self, output):
        self.output = output



import matplotlib.pyplot as plt
from sklearn.feature_selection import mutual_info_regression
import numpy as np
import pymc3 as pm
import scipy.stats as st
import pickle 
import datetime

class SimulationMetrics:
    """
    A method used to convert traces to something a data analyst can use to analyse
    """
    def __init__(self, traces=[], path=""):
        """
        Constructor for SimulationMetrics
        
        Parameters:
        ----------
        traces: List[Pymc3.trace]
            - A list of traces returned from simulation
        path: String
            - The path to a stored SimulationMetrics
        """
        if isinstance(traces, str):
            path = traces
            traces = []
        if len(traces):
            self.traces = traces
        else:
            self.traces = self.load_from_file(path)
        self.I = []
    
    def __str__(self):
        """
        Overrides the __str__ str method to print traces
        """
        return self.traces.__str__()

    def load_from_file(self, location):
        """
        A method which loads a pickled object as the trace

        Returns:
        ----------------
            - A "un"pickled object

        Parameters:
        ----------------
        Location: str
            - The location of the file
        """
        with open(location, "rb") as file:
            return pickle.load(file)

    def plot_mutual_bar(self, shift=0):
        """
        A method used in case of multiple parameters
        Is needed since the traces are than stored in a different sense

        Parameters
        ------------------
        Shift: int 
            - Since there are quite a large number of simulation a shift in the data can be needed

        """
        start, stop = 0,10
        executions = 5 #len(sm.traces)//2//10
        fig, ax = plt.subplots(executions,2, figsize=(20,18))
        for j in range(shift, shift+executions):
            for p in range(2):
                start = (p*10)+(j*20)
                stop = ((p*10)+10)+(j*20)
                traces = self.traces[start:stop]
                Is = {}
                labels = []
                for i in range(10):
                    trace = traces[i][0]
                    name = traces[i][2][0]
                    alice = trace[f"intDist_{i}"]
                    output = trace[f"Output_{i}"]
                    I = mutual_info_regression([[a] for a in alice], output, discrete_features=True)[0]
                    if name[0] in Is:
                        Is[name[0]].append(I)
                    else:
                        labels.append(name[0])
                        Is[name[0]] = [I]
                values = len(Is.keys())
                vals = [max(v[1]) for v in Is.items()]
                ax[j-shift][p].bar(labels, vals)
                ax[j-shift][p].set_ylim(0,7)
                ax[j-shift][p].set_title(f"Parameter {p} opposite was {self.traces[start+stop//2][2][1]}")
                ax[j-shift][p].set_ylabel("$I(X;Y)$")
        plt.tight_layout()
        plt.show()

    def plot_mutual_information(self, figsize=(16,8), as_bar=True):
        """
        Plots the mutual information as a graph for each distribution

        Parameters
        --------------
        figsize: Tuple<Int>
            - The size of the images
        as_bar: bool
            - Determines if the distribution should be a dot plot or a bar plot
        """
        I = self.mutual_information()
        size = len(I) if len(I) > 1 else 2
        plt.style.use('seaborn-darkgrid')
        _, ax = plt.subplots(2, 2,figsize=figsize)
        ylim = round(max(self.highest_leakage(head=1, verbose=0), key=lambda x: x[0])[0][0]+0.5)
        for pos, (axs, values) in enumerate(zip(ax.flatten(), I)):
            x = 0
            items = values.items()
            labels = []
            best_vals = []
            for k,v in items:
                if as_bar:
                    best = max(v, key=lambda x: x[0])
                    best_vals.append(best[0])
                else:
                    for value, info in v:
                        axs.plot([x], value, "x", label=str(info[1:]))
                best_info = round(max(v, key=lambda x: x[0])[0],2)
                axs.annotate(str(best_info), xy=(x, best_info), fontsize=16)
                if str(k) == "TruncatedNormal":
                    labels.append("Truncated \n Normal")
                else:
                    labels.append(str(k))
                x+=1
            if as_bar:
                axs.bar([i for i in range(len(best_vals))], best_vals)
            alice = {0: (0,100), 1: (0,300), 2: (0,10), 3: (0,1)}
            pi_pos = "A_{\pi_" + str(pos+1) + "}"
            title = f"$I(Y_{pos+1};{pi_pos})$ for parameter {pos+1} where ${pi_pos}$ ~ $U$ {alice[pos]}"
            ylabel = f"$I(Y_{pos+1};{pi_pos})$"
            axs.set_title(title, fontsize=16)
            axs.set_ylabel(ylabel, fontsize=14)
            axs.set_xlabel(f"Distributions", fontsize=14)
            axs.set_xticks(range(len(labels)))
            axs.set_xticklabels(labels, fontsize=14)
            axs.set_ylim(0,ylim)
            pos += 1
        plt.tight_layout()
        plt.show()

    def highest_leakage(self, head=1, verbose=1):
        """
        A method used to calculate the highest leakage grouped by each distribution

        Returns
        -----------
        List[Tuple[Float, Tuple[String, ]]]
            - Returns a list containing the mutual information next to the specific distribution

        Parameters
        -----------
        head: Int
            - Determines how many distributions are included
        verbose: int
            - Detmines if the distributions should be printed
        """
        if not len(self.I):
            self.mutual_information()
        best_vals = []
        for parameter_pos, l in enumerate(self.I):
            best_dist = []
            for k,v in l.items():
                best = max(v, key=lambda x: x[0])
                best_dist.append(best)
            best_dist = list(sorted(best_dist, key=lambda x: x[0], reverse=True))
            if verbose:
                print(f"The distribution that had the most leakage was {best_dist[:head]} for parameter {parameter_pos}")
            best_vals.append(best_dist[:head])
        return best_vals

    def save_to_file(self, location=""):
        """
        Save the particular trace to a file with the format:  Metrics-%Y-%m-%d-%H-%M-%S.priv

        Parameter:
        -----------
        location: string
            - The location in which the files should be saved
        """
        date = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        with open(location+f"metrics-{date}.priv", "wb") as file:
            pickle.dump(self.traces, file)

    def mutual_information(self):
        """
        Calculates the mutual information for each distribution and appends them to a global variable I

        Returns
        --------
        List[float, List[String,]]
            - A list of the mutual information paired with its respective distribution
        """
        _, names, _ = self.traces[0]
        size = len(names)
        mutual_information = [{} for _ in range(size)]
        for i in range(size):
            for trace, names, info in self.traces:
                discrete = ("int" in str(names[i]))
                alice = trace[names[i]]
                try:
                    output = trace["Output"]
                except:
                    pos = int(str(names[i]).split("_")[-1])
                    if pos < 10:
                        output = trace[f"Output_{pos}"]
                    else:
                        continue
                I_ao = mutual_info_regression([[j] for j in alice], output, discrete_features=discrete)[0]
                while (len(info) == 1):
                    info = info[0] # Used to unwrap the inner information in case of subtypes such as List[List[Tuple[...]]]
                if isinstance(info, tuple) or (isinstance(info, list) and isinstance(info[0], list)):
                    info = info[i]
                if info[0] in mutual_information[i]:
                    mutual_information[i][info[0]].append((I_ao,info))
                else:
                    mutual_information[i][info[0]] = [(I_ao, info)]
        self.I = mutual_information
        return mutual_information




class TheanoToken():
    """
    A enum-like class representing mappings from string values to theano types 
    """
    int_vector = "VectorI"
    float_vector = "VectorF"
    int_scalar = "int"
    float_scalar = "float"
    int_matrix = "MatrixI"
    float_matrix = "MatrixD"
    single_element_float_vector = 'Single_element_VectorF'
    single_element_int_vector = 'Single_element_VectorI'


    



import ast
import astor
import sys
import inspect
import os
import privugger.transformer.PyMC3.annotation_types as at
from privugger.transformer.PyMC3.theano_types import TheanoToken


class FunctionTypeDecorator(ast.NodeTransformer):

    function_name = None
    itypes = []
    otypes = []
    has_tuples = False
    has_list_tuples = False

 

    def __init__(self, name=None):
        self.function_name = name

    
    def get_function_def_ast(self, tree):
        
        for i in range(len(tree)):
            if(isinstance(tree[i], ast.FunctionDef)):
                return tree[i]
        raise TypeError("did not find any function definition in program")
    
    
    def get_function_return(self, body):
        
        for ast_node in body:
            if(isinstance(ast_node, ast.Return)):
                return ast_node
        raise TypeError("did not find any Return")

    def simple_method_wrap(self, program, name, args):
        
        arg_identifiers = []
        for a in args.args:
            arg_identifiers.append(a.arg)

        #func_returns = self.get_function_return(program.body)
        returns = ast.Return(value=ast.Call(args=[ast.arguments(args=arg_identifiers, defaults=[], vararg=None, kwarg=None)],func=ast.Name(id=name, ctx=ast.Load()), keywords=[]))
        
        new_function = ast.Module(body=[ast.FunctionDef(name='method', decorator_list=[], args=args, body=[program, returns])])
        return new_function
        #print(astor.to_source(new_function))
        #print(ast.dump(new_function))

    def lift(self, program, decorators):
        """
        This funtion provides another path to program lifting, when the decoration types are given directly. 
        The function lifts the program to be used within a pymc3 model.
         
        Parameters
        ------------
        program: path to program or a string of the entire program
        decorators: list of the decorator types

        Return
        -----------
        Python AST node with the lifted program
        
        """

        #NOTE This is for when the program is given as a path to the file
        if(isinstance(program, str)):
            file = open(program)
            tree = ast.parse(file.read())
            file.close()

        else:
            res = "".join(inspect.getsourcelines(program)[0])
            if "lambda" in res:
                res = res.replace(" ", "")
                start = res.find("lambda")
                end_draws = res.find("draws")
                end_chains = res.find("chains")
                end_cores = res.find("cores")

                #NOTE this is when we only use the default values
                if(end_draws == -1 and end_chains==-1 and end_cores == -1):

                    #TODO There are bugs if the program is defined elsewhere and ends with ")"
                    end = len(res)
                    
                    res = res[start:end-1].strip("\\n")

                else:
                    values = [end_draws, end_chains, end_cores]
                    #NOTE filter all the values that are not set 
                    positive_values = list(filter(lambda x: x != -1, values))
                    end = min(positive_values)
                    res = res[start:end-1].strip("\\n")
                
            if("lambda" == res[:6]):
                form = res.strip().split(":")
                res = f"def function({form[0][6:]}): \n return {form[1]}"

            elif("def" != res[:3]):
                raise TypeError("The program needs to be a path to a file, a lambda or a function")

            with open("temp.py", "w") as file:
                file.write(res)
            tree = ast.parse(open("temp.py").read())

            os.remove("temp.py")
        
        #if("lambda")
        
        
        
        function_def = self.get_function_def_ast(tree.body)

        node = self.create_decorated_function(function_def, decorators[0], decorators[1][0])

        if(isinstance(tree.body[0], ast.Import)):
            imports = tree.body[0]
            wrapped_node = self.simple_method_wrap(node, tree.body[1].name, tree.body[1].args)
            wrapped_node.body.insert(0, imports)
            
            
        else:
            wrapped_node = self.simple_method_wrap(node, tree.body[0].name, tree.body[0].args)
        
        return wrapped_node

    
    def translate_type(self, p_type):
        """

        Helper function to translate types from python types to theano tensor type
        Parameter: 
        string with the name of the type
        p_type: string

        Return:
        string with name of theano tensor type
        
        """
        if (p_type == 'float'):
            return 'dscalar'
        
        elif(p_type == 'int'):
            return 'lscalar'
        
        elif(p_type == 'VectorI'):
            return 'lvector'
        
        elif(p_type == 'VectorF'):
            return 'dvector'

        elif(p_type == 'MatrixI'):
            return 'lmatrix'

        elif(p_type== 'MatrixF'):
            return 'fmatrix'

        elif(p_type=='MatrixD'):
            return 'dmatrix'

        elif(p_type=='Single_element_VectorF'):
            return 'TensorType(\'float64\', (True,))'

        elif(p_type=='Single_element_VectorI'):
            return 'TensorType(\'int64\', (True,))'
        
        else:
            raise TypeError("Cannot translate type to any theano type")

    def from_python_to_theano_types(self, itypes, otype):

        """
        Function that transforms a list of itypes and a otype from python types to theano tensor types

        """
        theano_itypes = []
        for t in itypes:
            theano_itypes.append(self.translate_type(t))
        theano_otype = self.translate_type(otype)
        return (theano_itypes, theano_otype)



    def find_return_ast(self, body):
                
        """
        This function will find the return statements in original program

        """
        return_list = []
        for i in range(len(body)):
            if(isinstance(body[i], ast.Return)):
                return_list.append((body[i], i))
            else:
                pass
                #print(body[i])
        if(len(return_list) > 0):
            return return_list
        else:
           return None 

    def wrap_output_type(self, out_body, out_type):
        if((isinstance(out_body, ast.Name) or isinstance(out_body, ast.Compare) or isinstance(out_body, ast.ListComp) or isinstance(out_body, ast.BinOp) )):
            if(out_type == 'int'):
                o_attr = 'int64'
            #elif(out_type == 'float'):
             #   o_attr = 'float64'
            elif(out_type=='VectorI' or out_type == 'VectorF' or out_type == 'MatrixF' or 'MatrixI' or 'MatrixD' or 'float'):
                o_attr = 'array'
            return ast.Return(ast.Call(func=ast.Attribute(value=ast.Name(id='np', ctx=ast.Load()),attr=o_attr, ctx=ast.Load()), args=[out_body],keywords=[]))
        


        if( isinstance(out_body, ast.Constant)):
            if(out_type == 'int'):
                o_attr = 'int64'
            elif(out_type == 'float'):
                o_attr = 'float32'
            elif(out_type=='VectorI' or out_type == 'VectorF' or out_type == 'MatrixF' or 'MatrixI' or 'MatrixD'):
                o_attr = 'array'
            return ast.Return(ast.Call(func=ast.Attribute(value=ast.Name(id='np', ctx=ast.Load()),attr=o_attr, ctx=ast.Load()), args=[out_body],keywords=[]))
        
        if( isinstance(out_body, ast.IfExp)):
            if(out_type == 'int'):
                o_attr = 'int64'
            elif(out_type == 'float'):
                o_attr = 'float32'
            elif(out_type=='VectorI' or out_type == 'VectorF' or out_type == 'MatrixF' or 'MatrixI' or 'MatrixD'):
                o_attr = 'array'
            wrapped_body =  ast.Call(func=ast.Attribute(value=ast.Name(id='np', ctx=ast.Load()),attr=o_attr, ctx=ast.Load()), args=[out_body.body], keywords=[])
            wrapped_orelse =  ast.Call(func=ast.Attribute(value=ast.Name(id='np', ctx=ast.Load()),attr=o_attr, ctx=ast.Load()), args=[out_body.orelse], keywords=[])
            out_body.body = wrapped_body
            out_body.orelse = wrapped_orelse
            return (ast.Return(out_body))
        
        return (ast.Return(out_body))


    def create_decorated_function(self, node, itypes, otype):

        """
        This function decorates a FunctionDef node with tensor types

        """
        
        number_of_itypes = len(itypes) 
        
        theano_itypes,theano_otype =  self.from_python_to_theano_types(itypes, otype)

        ielts = []
        oelts = []
        for i in range(number_of_itypes):
            attr = ast.Attribute(value=ast.Name(id='tt', ctx=ast.Load()), attr=theano_itypes[i], ctx=ast.Load())
            ielts.append(attr) 
        
        oelts.append(ast.Attribute(value=ast.Name(id='tt', ctx=ast.Load()), attr=theano_otype, ctx=ast.Load()))
        
        theano_keywords = [ast.keyword(arg='itypes', value=ast.List(elts=ielts)), ast.keyword(arg='otypes', value=ast.List(elts=oelts))]

        theano_decorator_list = [ast.Call(func=ast.Attribute(value=ast.Attribute(
        value=ast.Attribute(value=ast.Name(id='theano', ctx=ast.Load()), attr='compile', ctx=ast.Load()), attr='ops', ctx=ast.Load()), attr='as_op', ctx=ast.Load()), args=[], keywords=theano_keywords)]
        
        #return_body, index = self.find_return_ast(node.body)
        return_list = self.find_return_ast(node.body)
        if(return_list is not None):
            for r in return_list:
                wrapped_return_body = self.wrap_output_type(r[0].value, otype)
                node.body[r[1]] = wrapped_return_body
        node.decorator_list = theano_decorator_list
        return node 


    def get_next_annotation(self, arg):
        
        if(isinstance(arg, ast.List)):
            l = at.List()
            l.a_type = self.get_next_annotation(arg.elts[0])
            return (l)

        elif(isinstance(arg, ast.Tuple)):
    
            t = at.Tuple()
            t.length = len(arg.elts) 
            for i in range(t.length):
                t.a_type.append(self.get_next_annotation(arg.elts[i]))
            return (t)

        
        elif(isinstance(arg, ast.Subscript)):

            if(arg.value.id == "List"):
                l = at.List()
                l.a_type = self.get_next_annotation(arg.slice.value)
                return l
            else:
                return self.get_next_annotation(arg.slice.value)
          
        else:
            if(arg.id == 'int'):
                return (at.Int())
            if(arg.id == 'float'):
                return (at.Float())
            else:
        
                raise TyperError("Seems that there is a problem with the annotation")

    def visit_FunctionDef(self, node):
        """
        The main traversal function that visits all the FunctionDef nodes and returns a decorated FunctionDef

        """
        if(isinstance(node, ast.FunctionDef)):
            if(node.name != self.function_name):
                return None
        
        itypes = []
        otype = None
    
        for a in node.args.args:
            #annotation = self.get_next_annotation_rec(a.annotation, "")
            annotation_type = self.get_next_annotation(a.annotation)
            if(type(annotation_type) == at.List):
                if(type(annotation_type.a_type) == at.Int):
                    #itypes.append("VectorI")
                    itypes.append(TheanoToken.int_vector)
                elif(type(annotation_type.a_type) == at.Float):
                    #itypes.append("VectorF")
                    itypes.append(TheanoToken.float_vector)
                elif(type(annotation_type.a_type) == at.Tuple):
                    self.has_list_tuples = True
                    for t in annotation_type.a_type.a_type:
                        if(type(t) == at.Int):
                            itypes.append(TheanoToken.int_vector)
                            #itypes.append("VectorI")
                        if(type(t) == at.Float):
                            #itypes.append("VectorF")
                            itypes.append(TheanoToken.float_vector)
        
            elif(type(annotation_type) == at.Int):
                #itypes.append("int")
                itypes.append(TheanoToken.int_scalar)
            
            elif(type(annotation_type) == at.Float):
                #itypes.append("float")
                itypes.append(TheanoToken.float_scalar)
            
            elif(type(annotation_type) == at.Tuple):
                self.has_tuples = True
                for t in annotation_type.a_type:
                    if(type(t) == at.Int):
                        #itypes.append("VectorI")
                        itypes.append(TheanoToken.int_vector)
                    if(type(t) == at.Float):
                            #itypes.append("VectorF")
                            itypes.append(TheanoToken.float_vector)
         
            else:
                raise TypeError("Seems that the annotations are wrong")
            
        

        out_annotation_type = self.get_next_annotation(node.returns)
        if(type(out_annotation_type) == at.List):
            if(type(out_annotation_type.a_type) == at.List):
                if(out_annotation_type.a_type.a_type == at.Int):
                    outype = TheanoTokoen.int_matrix
                else:
                    otype = TheanoToken.float_matrix
            elif(type(out_annotation_type.a_type) == at.Tuple):
                all_ints = all(isinstance(t, at.Int) for t in out_annotation_type.a_type.a_type)
                if all_ints:
                    otype = TheanoToken.int_matrix
                else:
                    otype = TheanoToken.float_matrix
            elif(type(out_annotation_type.a_type) == at.Int):
                otype = TheanoToken.int_scalar
            elif(type(out_annotation_type.a_type) == at.Float):
                otype = TheanoToken.float_scalar
        elif(type(out_annotation_type) == at.Int):
            otype = (TheanoToken.int_scalar)
            
        elif(type(out_annotation_type) == at.Float):
            otype = (TheanoToken.float_scalar)
            
        elif(type(out_annotation_type) == at.Tuple):
            all_ints = all(isinstance(t, at.Int) for t in out_annotation_type.a_type)
            if(all_ints):
                otype = TheanoToken.int_matrix
            else:
                otype = TheanoToken.float_matrix
         
        else:
            raise TypeError("Seems that the annotations are wrong")

        self.itypes = itypes
        self.otypes = otype
    
        decorated_func = self.create_decorated_function(node, itypes, otype)    
        return ((decorated_func))




    def find_function_def_idx(self, body):
        iterable_list = body.body
        for i in range(len(iterable_list)):
            if( isinstance(iterable_list[i], ast.FunctionDef)):
                return (i)
        return (-1)

    
    def wrap_with_theano_import(self, program):

        theano_import = ast.Import(names=[ast.alias(name='theano', asname=None)])
        theano_tensor_import = ast.Import(names=[ast.alias(name='theano.tensor', asname='tt')])
        numpy_import = ast.Import(names=[ast.alias(name='numpy', asname='np')])
        new_program = ast.Module(body=[theano_import, theano_tensor_import,numpy_import,  program])
        
        return new_program

    def wrap_with_imports(self, program):
        theano_import = ast.Import(names=[ast.alias(name='theano', asname=None)])
        theano_tensor_import = ast.Import(names=[ast.alias(name='theano.tensor', asname='tt')])
        numpy_import = ast.Import(names=[ast.alias(name='numpy', asname='np')])
        typing_import = ast.ImportFrom(level=0, module='typing', names=[ast.alias(name='List', asname=None), (ast.alias(name='Tuple', asname=None))])
        functools_import = ast.ImportFrom(level=0, module='functools', names=[ast.alias(name='reduce', asname=None)])


        body_with_needed_imports = [theano_import, theano_tensor_import, numpy_import,functools_import, typing_import]+ program.body 
        program.body = body_with_needed_imports

        return program


    def construct_python_args(self):

        new_args = []

        for i in range(len(self.itypes)):
            new_annotation = None

            if(self.itypes[i] == 'VectorI'):
                new_annotation = ast.Subscript(value=ast.Name(id = "List",ctx = ast.Load()),slice=ast.Name(id='int', ctx=ast.Load()))
            if(self.itypes[i]  == 'VectorF'):
                new_annotation = ast.Subscript(value=ast.Name(id = "List",ctx = ast.Load()),slice=ast.Name(id='float', ctx=ast.Load()))
            if(self.itypes[i]  == 'int'):
                new_annotation = ast.Name(id='int', cts=ast.Load())
            if(self.itypes[i]  == 'float'):
                new_annotation = ast.Name(id='float', cts=ast.Load())
            if(self.itypes[i]  == 'MatrixI'):
                pass
            if(self.itypes[i]  == 'MatrixF'):
                pass
            new_args.append(ast.arg(arg=f'arg_{i}',annotation=new_annotation))

        return ast.arguments(args=new_args, defaults=[], vararg=None, kwarg=None)

    def construct_python_body(self, body, new_args, new_args_idx, original_name):
        
        value_names = []

        for i in new_args_idx:
            value_names.append(ast.Name(id = new_args.args[i].arg, ctx=ast.Load()))
        
        body.args = new_args
        if(self.has_tuples and not self.has_list_tuples):
            temp = [ast.Assign(targets=[ast.Name(id=original_name, ctx=ast.Load())], value=ast.Call(args=value_names, func=ast.Name(id='tuple', ctx=ast.Load()), keywords=[]))] + body.body
            body.body = temp    
        if(self.has_list_tuples):
            temp = [ast.Assign(targets=
                    [ast.Name(id=original_name, ctx=ast.Load())],
                    value=ast.Call(args=[ast.Call(args=value_names, 
                    func=ast.Name(id='zip', ctx=ast.Load()), keywords=[])], func=ast.Name(id='list', ctx=ast.Load()), keywords=[]))] + body.body
            body.body = temp

        return body

    def wrap_program_with_signature(self, program):

        idx = self.find_function_def_idx(program)
        if(idx == -1):
            raise NotImplementedError()
        
        func_name = program.body[idx].name
        func_args = program.body[idx].args

        if(self.has_tuples or self.has_list_tuples):

            new_args = self.construct_python_args()
            new_args_idx = []
            original_name = None
            for i in range(len(new_args.args)):
                new = new_args.args[i].annotation
                if(not i >= len(func_args.args)):
                    old = func_args.args[i].annotation
                    if(isinstance(old, ast.Subscript) and not isinstance(new, ast.Subscript)):
                        new_args_idx.append(i)
                    
                    if(isinstance(old, ast.Subscript)):
                        if(isinstance(old.slice.value, ast.Tuple)):
                            new_args_idx.append(i)
                            original_name = func_args.args[i].arg
                        elif(old.slice.value.value.id == "Tuple"):
                            new_args_idx.append(i)
                            original_name = func_args.args[i].arg

                else:
                    new_args_idx.append(i)
                    # print(new_args_idx)


            new_body = self.construct_python_body(program.body[idx], new_args, new_args_idx, original_name)
        else:
            new_args = func_args
            new_body = program.body[idx]

  
        #Check every arguments, to see if it contains a tuple or a list of tuples
        # for i in range(len(func_args.args)):
        #     idx_has_tuple_arguments.append((i, has_tuple(func_args.args[i].annotation), has_list(func_args.args[i].annotation)))
        # print(idx_has_tuple_arguments)

        func_returns = program.body[idx].returns
        
        arg_identifiers = []
        for a in new_args.args:
            arg_identifiers.append(a.arg)
               
    
        returns = ast.Return(value=ast.Call(args=[ast.arguments(args=arg_identifiers, defaults=[], vararg=None, kwarg=None)], func=ast.Name(id=func_name, ctx=ast.Load()), keywords=[]))

        #program.body[idx]
        new_function = ast.Module(body=[ast.FunctionDef(name='method', decorator_list=[], args=new_args, body=[new_body, returns ], returns=func_returns)])

        return (new_function)


def load(path, function):
 
    tree =ast.parse(open(path).read())

    #print(ast.dump(tree))
    ftp = FunctionTypeDecorator(function)

    new_program = ftp.visit(tree)

    new_program_with_outer_function = ftp.wrap_program_with_signature(new_program)

    new_program_with_imports = ftp.wrap_with_imports(new_program_with_outer_function)
    #TheanoImport().visit(new_program_with_outer_function)
    print(astor.to_source(new_program_with_imports))


    return new_program_with_imports






"""
Probability distributions generators
"""
import pymc3 as pm
import pymc3.distributions as dist
from privugger.attacker.distributions import *
from hypothesis import strategies as st
import numpy as np
import scipy

def IntList(name, data, length=1, possible_dist=POSSIBLE_INTS, ranges=(0, np.inf)):
    """
    Generates a list of probabilistics distributions to mimic all possible int values
    
    Returns: Tuple[List[Pymc3.distributions], List[Tuple[String, ]]
    ----------------------------------------------------------------
    Returns a tuple containing a list of pymc3 distributions and a list containing information about each distributions
    
    Parameters:
    ------------
    name: String
        - The general name the list of distributions should take. In case of multiple distributions than names will have appended # within name.
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    length: Int
        - The length of the list
    possible_dist: List[Int]
        - A list of ints to be chosen from Privugger.distributions, indicating which distributions to choose from.
    """
    rand = data.draw(st.randoms(use_true_random=True))
    use_multiple_dist = rand.choice([0])
    if use_multiple_dist and len(possible_dist) < 0:
        dist, info = tuple(zip(*[IntGenerator(data=data, name=f"{name}{i}", shape=1) for i in range(length)]))
        return (dist, info)
    else:
        return IntGenerator(data=data, name=name, shape=length, ranges=ranges)

def FloatList(name, data, length=1, possible_dist=POSSIBLE_FLOATS, ranges=(-np.inf, np.inf)):
    """
    Generates a list of probabilistics distributions to mimic all possible float values
    
    Returns: Tuple[List[Pymc3.distributions], List[Tuple[String, ]]
    ----------------------------------------------------------------------
    Returns a tuple containing a list of pymc3 distributions and a list containing information about each distributions

    Parameters:
    -----------
    name: String
        - Name of the distributions. In case of multiple dist the names will also have appended the #
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    length: Int
        - The length of the list
    possible_dist: List[Int]
        - A list of ints to be chosen from privugger.distributions, indicating which distributions to choose from
    """
    rand = data.draw(st.randoms(use_true_random=True))
    use_same_shape = rand.choice([1])
    if use_same_shape:
        return FloatGenerator(name, data, possible_dist=[TRUNCATED_NORMAL], shape=length, ranges=ranges)
    else:
        dist, info = tuple(zip(*[FloatGenerator(name+str(i), data, possible_dist=possible_dist, ranges=ranges) for i in range(length)]))
        return (dist, info)


def IntGenerator(data, name, possible_dist = POSSIBLE_INTS, shape=1, ranges=(0, np.inf)):
    """
    A method for generating a single probabilistic distributions to mimic int distribution

    Returns: Tuple[Pymc3.distributions, Tuple[String, ]]
    ----------
        - Returns a tuple containint the distribution and a tuple with information about the distribution

    Parameters
    ----------
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    name: str
        - The name of the ditributions
    possible_dist: List[ints]
        - A list of the desired privugger.distributions to be chosen random from
    shape: Int
        - The dimensionality of the distribution
    """
    if ranges[0] < 0 or ranges[0] >= ranges[1]:
        raise ValueError("The ranges has to be greater than or equal to 0 and in increasing order. E.g. (0,100)")
    rand = data.draw(st.randoms(use_true_random=True))
    if ranges[1] > 1 and BERNOULLI in possible_dist:
        possible_dist.remove(BERNOULLI)
    dist = rand.choice(possible_dist)
    if dist == BINOMIAL:
        return Binomial(data=data, name=name, shape=shape, ranges=ranges)
    elif dist == BERNOULLI:
        return Bernoulli(data=data, name=name, shape=shape, ranges=ranges) 
    elif dist == GEOMETRIC:
        return Geometric(data=data, name=name, shape=shape, ranges=ranges)
    elif dist == BETA_BINOMIAL:
        return BetaBinomial(name=name, data=data, shape=shape, ranges=ranges)
    elif dist == POISSON:
        return Poisson(name=name, data=data, shape=shape, ranges=ranges)
    elif dist == DISCRETE_UNIFORM:
        return DiscreteUniform(name=name, data=data, shape=shape, ranges=ranges)
    else:
        raise ValueError("The possible distribution is not supported for Int Generators")

def FloatGenerator(name, data, possible_dist = POSSIBLE_FLOATS, shape=1, ranges=(-np.inf, np.inf)):
    """
    A method for generating a single distributions to represent float data

    Returns: Tuple[Pymc3.distributions, Tuple[String, ]]
    ----------
        - Returns a tuple containint the distribution and a tuple with information about the distribution

    Parameters
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    possible_dist: List[ints]
        - A list of the desired privugger.distributions to be chosen random from
    shape: Int
        - The dimensionality of the distribution
    """
    rand = data.draw(st.randoms(use_true_random=True))
    dist = rand.choice(possible_dist)
    if dist == NORMAL:
        return Normal(data=data, name=name, shape=shape, ranges=ranges)
    elif dist == UNIFORM:
        return Uniform(data=data, name=name, shape=shape, ranges=ranges)
    elif dist == TRUNCATED_NORMAL:
        return TruncatedNormal(name=name, data=data, shape=shape)
    elif dist == BETA:
        return Beta(name=name,data=data, shape=shape)
    elif dist == EXPONENTIAL:
        return Exponential(name=name, data=data, shape=shape)
    elif dist == LAPLACE:
        return Laplace(name=name, data=data, shape=shape)
    elif dist == STUDENT_T:
        return StudentT(name=name, data=data, shape=shape)
    elif dist == CAUCHY:
        return Cauchy(name=name, data=data, shape=shape)
    elif dist == GAMMA:
        return Gamma(name=name, data=data, shape=shape)
    else:
        raise ValueError("The possible distribution is not supported for Int Generators")

# Int Distributions

def Binomial(data, name, shape=1, ranges=(1, np.inf)):
    """
    Constructs a binomial distributions with RV = X ~ Binomial(n,p)

    Returns: Tuple[Pymc3.distributions.Binomial, Tuple[String, int, float]]
    ----------
        - Returns a tuple with binomial distributions paired with the name, numper of test and probability
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    l,h = ranges
    mean = lambda n,p: n*p
    ints = st.integers(min_value=l, max_value=10000)
    probability = st.floats(min_value=0.0010004043579101562, max_value=0.990234375, allow_infinity=False, allow_nan=False, width=16)
    n,p = data.draw(st.tuples(ints,probability).map(sorted).filter(lambda x: l <= mean(x[0],x[1]) <= h))
    if shape > 1:
        a = dist.Binomial(name=name, n=n, p=p, shape=shape)
    else:
        a = dist.Binomial(name=name, n=n, p=p)
    b = ["Binomial", n,p]
    return (a,b)


def Bernoulli(data, name, shape=1, ranges=(0,1)):
    """
    Constructs a bernoulli distributions with RV = X ~ Bernoulli(p)

    Returns: Tuple[Pymc3.distributions.Bernoulli, Tuple[String, float]]
    ----------
        - Returns a tuple with Bernoulli distributions paired with the name and probability
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    probability = st.floats(min_value=0.0010004043579101562, max_value=0.990234375,allow_infinity=False, allow_nan=False,width=16)
    p = data.draw(probability)
    if shape > 1:
        a = dist.Bernoulli(name=name, p=p, shape=shape)
    else:
        a = dist.Bernoulli(name=name, p=p)
    b = ["Bernoulli", p]
    return (a,b)
    

def Geometric(data, name, shape=1, ranges=(1, np.inf)):
    """
    Constructs a geometric distributions with RV = X ~ Geometric(p)

    Returns: Tuple[Pymc3.distributions.Geometric, Tuple[String, float]]
    ----------
        - Returns a tuple with geometric distributions paired with the name and probability
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    l,h = ranges
    mean = lambda p: 1/p
    probability = st.floats(min_value=0.0010004043579101562, max_value=0.990234375,allow_infinity=False, allow_nan=False,width=16).filter(lambda x: l <= mean(x) <= h)
    p = data.draw(probability)
    if shape > 1:
        a = dist.Geometric(name=name, p=p, shape=shape)
    else:
        a = dist.Geometric(name=name, p=p)
    b = ["Geometric", p]
    return (a,b)


def BetaBinomial(data, name, shape=1, ranges=(1, np.inf)):
    """
    Constructs a BetaBinomial distributions with RV = X ~ BetaBinomial(n, a, ß)

    Returns: Tuple[Pymc3.distributions.BetaBinomial, Tuple[String, int, float, float]]
    ----------
        - Returns a tuple with distributions paired with the [name, number, alpha, beta]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    l,h = ranges
    mean = lambda n,a,b: (n*a)/(a+b)
    ints = st.integers(min_value=l, max_value=10000)
    positive_float = st.floats(min_value=0.0999755859375, max_value=10000,allow_infinity=False, allow_nan=False,width=16)
    tuples = st.tuples(ints,positive_float,positive_float).map(sorted).filter(lambda x: l <= mean(x[0],x[1],x[2]) <= h)
    n = data.draw(ints)
    alpha = data.draw(positive_float)
    beta = data.draw(positive_float)
    if shape > 1:
        a = dist.BetaBinomial(name, alpha=alpha, beta=beta, n=n, shape=shape)
    else:
        a = dist.BetaBinomial(name, alpha=alpha, beta=beta, n=n)
    b = ["BetaBinomial", n, alpha, beta]
    return (a,b)


def Poisson(data, name, shape=1, ranges=(0, np.inf)):
    """
    Constructs a Poisson distributions with RV = X ~ Poisson(µ)

    Returns: Tuple[Pymc3.distributions.Poisson, Tuple[String, float]]
    ----------
        - Returns a tuple with distributions paired with the [name, mu]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    l,h = ranges
    non_negativ_float = st.floats(min_value=l, max_value=10000,allow_infinity=False, allow_nan=False,width=16).filter(lambda x: l <= x <= h)
    mu = data.draw(non_negativ_float)
    if shape > 1:
        a = dist.Poisson(name, mu=mu, shape=shape)
    else:
        a = dist.Poisson(name, mu=mu)
    b = ["Poisson", mu]
    return (a,b) 


def DiscreteUniform(data, name, ranges=(-np.inf, np.inf), shape=1):
    """
    Constructs a DiscreteUniform distributions with RV = X ~ DiscreteUniform(l,u)

    Returns: Tuple[Pymc3.distributions.DiscreteUniform, Tuple[String, int, int]]
    ----------
        - Returns a tuple with distributions paired with the [name, lower, upper]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    low, high = min(ranges), max(ranges)
    low = low if low != -np.inf else -1000
    high = high if high != np.inf else 1000
    values = st.integers(min_value=low, max_value=high)
    mean = lambda l,h: (h+l)/2
    size = (st.tuples(values, values)
                .map(sorted)
                .filter(lambda x: x[0] < x[1] and  low <= mean(x[0],x[1]) <= high))
    lower, upper = data.draw(size)
    a = dist.DiscreteUniform(name, lower, upper)
    b = ["DiscreteUniform", lower, upper]
    return (a,b)

# Float Distributions
def Normal(data, name, shape=1, ranges=(0, 100)):
    """
    Constructs a Normal distributions with RV = X ~ Normal(µ,sigma)

    Returns: Tuple[Pymc3.distributions.Normal, Tuple[String, float, float]]
    ----------
        - Returns a tuple with distributions paired with the [name, mu, sigma]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    range: Tuple[int] 
        - The possible values that the PMF can mimic
    """
    low, high = ranges
    low = low if low != -np.inf else -1000
    high = high if high != np.inf else 1000

    floats = st.floats(allow_infinity=False, allow_nan=False, min_value=low, max_value=high,width=16)
    positive_floats = st.floats(min_value=0.0999755859375,allow_infinity=False, allow_nan=False,max_value=high-low,width=16)

    mean = lambda mu, sigma: mu
    values = st.tuples(floats, positive_floats).filter(lambda x: low <= mean(x[0], x[1]) <= high)
    mu, sigma = data.draw(values)
    a = dist.Normal(name=name, mu=mu, sigma=sigma, shape=shape)
    b = ["Normal", mu,sigma]
    return (a,b)


def Uniform(data, name, shape=1, ranges=(0,100)):
    """
    Constructs a Uniform distributions with RV = X ~ Uniform(l,u)

    Returns: Tuple[Pymc3.distributions.Uniform, Tuple[String, int, int]]
    ----------
        - Returns a tuple with distributions paired with the [name, lower, upper]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    low, high = ranges
    low = low if low != -np.inf else -1000
    high = high if high != np.inf else 1000
    floats = st.floats(allow_infinity=False, allow_nan=False, min_value=low, max_value=high,width=16)
    mean = lambda a,b: (a+b)/2
    size = (st.tuples(floats, floats)
                .map(sorted)
                .filter(lambda x: x[0] < x[1] and low <= mean(x[0],x[1]) <= high))
    lower, upper = data.draw(size)
    a = dist.Uniform(name, lower=lower, upper=upper, shape=shape)
    b = ["Uniform", lower, upper]
    return (a,b)


def TruncatedNormal(data, name, shape=1, ranges=(-np.inf, np.inf)):
    """
    Constructs a TruncatedNormal distributions with RV = X ~ TruncatedNormal(mu, sigma, lower, upper)

    Returns: Tuple[Pymc3.distributions.TruncatedNormal, Tuple[String, float, float, int, int]]
    ----------
        - Returns a tuple with distributions paired with the [name, mu, sigma, lower, upper]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    low, high = ranges
    low = low if low != -np.inf else -1000
    high = high if high != np.inf else 1000
    floats = st.floats(allow_infinity=False, allow_nan=False, min_value=low, max_value=high,width=16).filter(lambda x: low <= x <= high)
    positive_floats = st.floats(min_value=0.0999755859375,allow_infinity=False, allow_nan=False,max_value=high-low,width=16)
    size = (st.tuples(st.integers(min_value=low, max_value=high), st.integers(min_value=low, max_value=high))
                .map(sorted)
                .filter(lambda x: x[0] < x[1]))
    mu = data.draw(floats)
    sigma = data.draw(positive_floats)
    lower, upper = data.draw(size)
    a = dist.TruncatedNormal(name, mu=mu, sigma=sigma, lower=lower, upper=upper, shape=shape)
    b = ["TruncatedNormal", mu, sigma, lower, upper]
    return (a,b)


def Beta(data, name, shape=1):
    """
    Constructs a Beta distributions with RV = X ~ Beta(alpha, beta)

    Returns: Tuple[Pymc3.distributions.Beta, Tuple[String, float, float]]
    ----------
        - Returns a tuple with distributions paired with the [name, alpha, beta]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    positive_floats = st.floats(min_value=0.0999755859375,allow_infinity=False, allow_nan=False,max_value=40,width=16)
    alpha = data.draw(positive_floats)
    beta = data.draw(positive_floats)
    a = dist.Beta(name, alpha=alpha, beta=beta, shape=shape)
    b = ["Beta", alpha, beta]
    return (a,b)


def Exponential(data, name, shape=1, ranges=(-np.inf, np.inf)):
    """
    Constructs a Exponential distributions with RV = X ~ Exponential(lambda)

    Returns: Tuple[Pymc3.distributions.Exponential, Tuple[String, float]]
    ----------
        - Returns a tuple with distributions paired with the [name, lambda]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    low, high = ranges
    low = low if low != -np.inf else 0
    high = high if high != np.inf else 1000
        
    positive_floats = st.floats(min_value=0.0999755859375,allow_infinity=False, allow_nan=False,max_value=50,width=16).filter(lambda x: low <= 1/x <= high)
    lam = data.draw(positive_floats)
    a = dist.Exponential(name, lam, shape=shape)
    if not low:
        shift = pm.Deterministic(name+"_shifted", a+low)
        b = ["Exponential", lam]
        return (a,b)
    b = ["Exponential", lam]
    return (a,b)


def Laplace(data, name, shape=1, ranges=(-np.inf, np.inf)):
    """
    Constructs a Laplace distributions with RV = X ~ Laplace(mu, b)

    Returns: Tuple[Pymc3.distributions.Laplace, Tuple[String, float, float]]
    ----------
        - Returns a tuple with distributions paired with the [name, mu, b]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    low, high = ranges
    low = low if low != -np.inf else -1000
    high = high if high != np.inf else 1000
    floats = st.floats(allow_infinity=False, allow_nan=False, min_value=low, max_value=high,width=16).filter(lambda x: low <= x <= high)
    positive_floats = st.floats(min_value=0.0999755859375,allow_infinity=False, allow_nan=False,max_value=50,width=16)
    mu = data.draw(floats)
    bi = data.draw(positive_floats)
    a = dist.Laplace(name, mu=mu, b=bi, shape=shape)
    b = ["Laplace", mu, bi]
    return (a,b)


def StudentT(data, name, shape=1, ranges=(-np.inf, np.inf)):
    """
    Constructs a StudentT distributions with RV = X ~ StudentT(nu, mu, sigma)

    Returns: Tuple[Pymc3.distributions.StudentT, Tuple[String, float, float, float]]
    ----------
        - Returns a tuple with distributions paired with the [name, nu, mu, sigma]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    low, high = ranges
    low = low if low != -np.inf else -1000
    high = high if high != np.inf else 1000
    positive_floats = st.floats(min_value=0.0999755859375,allow_infinity=False, allow_nan=False,max_value=high-low,width=16)
    floats = st.floats(allow_infinity=False, allow_nan=False, min_value=low, max_value=high,width=16)
    nu = data.draw(positive_floats)
    mu = data.draw(floats)
    sigma = data.draw(positive_floats)
    a = dist.StudentT(name, nu=nu, mu=mu, sigma=sigma, shape=shape)
    b = ["StudentT", nu, mu, sigma]
    return (a,b)


def Cauchy(data, name, shape=1, ranges=(-np.inf, np.inf)):
    """
    Constructs a Cauchy distributions with RV = X ~ Cauchy(alpha, beta)

    Returns: Tuple[Pymc3.distributions.Cauchy, Tuple[String, float, float]]
    ----------
        - Returns a tuple with distributions paired with the [name, alpha, beta]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    low, high = ranges
    low = low if low != -np.inf else -1000
    high = high if high != np.inf else 1000
    positive_floats = st.floats(min_value=0.0999755859375,allow_infinity=False, allow_nan=False,max_value=10,width=16)
    floats = st.floats(allow_infinity=False, allow_nan=False, min_value=low, max_value=high,width=16)
    alpha = data.draw(floats)
    beta = data.draw(positive_floats)
    a = dist.Cauchy(name, alpha=alpha, beta=beta, shape=shape)
    b = ["Cauchy", alpha, beta]
    return (a,b)


def Gamma(data, name, shape=1):
    """
    Constructs a Gamma distributions with RV = X ~ Gamma(alpha, beta)

    Returns: Tuple[Pymc3.distributions.Gamma, Tuple[String, float, float]]
    ----------
        - Returns a tuple with distributions paired with the [name, alpha, beta]
    
    Parameters:
    ----------
    name: str
        - The name of the ditributions
    data: hypothesis.data
        - The hypothesis data used to draw the distributions
    shape: int
        - The dimensionality of the distribution
    """
    positive_floats = st.floats(min_value=0.0999755859375,allow_infinity=False, allow_nan=False,max_value=40,width=16)
    values = st.tuples(positive_floats, positive_floats).map(sorted).filter(lambda x: low <= (x[0]/x[1]) <= high)
    alpha, beta = data.draw(values)
    a = dist.Gamma(name, alpha, beta, shape=shape)
    b = ["Gamma", alpha, beta]
    return (a,b)


"Mutual information estimators"

import numpy as np
from sklearn.feature_selection import mutual_info_regression


def mi_sklearn(
        trace,
        var_names=[],
        disc_features=False,
        log2=True,
        n_neigh=20,
        input_inferencedata=True
):
    """
    Binding for estimating mutual information using the `mutual_info_regression` function from `sklearn`.

    Concretely, this function computes mutual information as I(var_names[0]; var_names[1]).      
   

    Parameters
    ----------
    trace : az.InferenceData object
        InferenceData object containing the samples for variables of interest
    var_names: String \times String
        String array with two elements indicating the variables used to compute mutual information.
        Note that the array must contain exactly two elements.
    disc_features : bool
        Boolean indicating whether the domain of the secret is discerete. Default False.
    log2: bool
        Result in log2. Default True. If False, the result is in natural logarithm.
    n_neigh: int
        Number of neighbours used by de estimator. Default 20.

    Returns
    -------
    result : float
        Mutual information in log2 or ln.
    """
    trace1 = trace2 = []
    if input_inferencedata:
        assert len(var_names)==2, "var_names must contain exactly two elements"
        trace1 = trace.posterior[var_names[0]].values.flatten().reshape(-1,1)
        trace2 = trace.posterior[var_names[1]].values.flatten()
    else:
        assert len(trace) == 2, "trace must containt two subtraces for each of each random variables to compute mutual information"
        trace1 = trace[0].reshape(-1,1)
        trace2 = trace[1]

    mi_nat = mutual_info_regression(trace1,
                                    trace2,
                                    discrete_features=disc_features,
                                    n_neighbors=n_neigh)
    result = mi_nat/np.log(2) if log2 else mi_nat
    return result



from setuptools import setup, find_packages


#
#NOTE: whenever we publish a new version come in here and bump up the version number first!
# Our version number follows the convention of 3 numbers: MAJOR (new api), MINOR (added functionality), PATCH (bug fix)
#
#pip install -e . to test locally before publishing
#build command: python3 setup.py bdist_wheel sdist
#push to PyPi command: twine upload dist/*
#push to testPyPi command: twine upload --repository testpypi dist/*


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


    
setup(
    name='privugger',
    version='0.0.9',
    license='Apache license',
    author='Raúl Pardo, Mathias Valdbjørn Jørgensen, and Rasmus Carl Rønneberg',
    author_email='raup@itu.dk',
    description='Privacy risk analysis library for Python programs.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/itu-square/privugger',
    #package_dir={"": "privugger"},
    packages=find_packages(),
)



import pymc3 as pm
from scipy import stats as st
from abc import abstractmethod
"""
By specifying our own interface for distributions we could ideally hide which specific backend is used to model the distributions
"""

class Continuous():

    """
    An abstract class representing all the continuous distributions that privugger supports
    """
    @abstractmethod
    def pymc3_dist(self, name, hypers):
        return None
	
    @abstractmethod
    def get_params(self):
    	return None
    	
    	
    @abstractmethod
    def scipy_dist(self, name):
        return None

__all__ = [
    "Uniform",
    "Normal",
    "Exponential",
    "Beta"
]


class Uniform(Continuous):
    """
        
    Class for the Uniform distribution 

    Attributes 
    --------------

    name : String of the name of the random variable
        
    lower : int for the lower bound. Default: 0
        
    upper : int for the upper bound. Default: 1
        
    num_elements : int specifying number of RV's
        
    is_hyper_param : Boolean specifying if this RV is used as a hyper parameter. Default: False    
       
    """
        
    
    def __init__(self,name, lower=0, upper=1, num_elements=-1, is_hyper_param=False):


        self.lower = lower
        self.name = name
        self.upper = upper
        self.num_elements = num_elements
        self.is_hyper_param = is_hyper_param
    
    def pymc3_dist(self, name, hypers):
        

        lower = self.lower
        upper = self.upper
        if(len(hypers) == 1):
                hyper_dist = hypers[0][0]
                hyper_name = hypers[0][1]
                idx = hypers[0][2]
                if(idx == 0):
                    lower = hyper_dist.pymc3_dist(hyper_name, [])
                else:
                    upper = hyper_dist.pymc3_dist(hyper_name, [])
        elif(len(hypers) == 2):
                hyper_dist_1 = hypers[0][0]
                hyper_name_1 = hypers[0][1]
                hyper_dist_2 = hypers[1][0]
                hyper_name_2 = hypers[1][1]

                lower = hyper_dist_1.pymc3_dist(hyper_name_1, [])
                upper = hyper_dist_2.pymc3_dist(hyper_name_2, [])
        if(self.num_elements==-1):
            return pm.Uniform(name, lower=lower, upper=upper)
        else:
            
            return pm.Uniform(name, lower=lower, upper=upper, shape=self.num_elements)
	
    def get_params(self):
    	return [self.lower, self.upper]
    
    def scipy_dist(self, name):
        dist = (lambda siz : st.uniform(self.lower, self.upper-self.lower).rvs(siz)) if self.num_elements == -1 else (lambda siz: st.uniform(self.lower, self.upper-self.lower).rvs((self.num_elements, siz)))
        return name,dist



class Normal(Continuous):

    """
    Class for the Gaussian distribution 
    
    Attributes 
    -----------
    name: String of the name of the random variable

    mu: value for the mean of the distribution. Default: 0

    std: value for the standard deviation. Default: 1

    num_elements: int specifying number of RV's

    is_hyper_param: Boolean specifying if this RV is used as a hyper parameter. Default: False
    """

    def __init__(self,name, mu=0, std=1, num_elements=-1, is_hyper_param=False):

        self.mu = mu
        self.name = name
        self.std = std
        self.num_elements = num_elements
        self.is_hyper_param = is_hyper_param
   
    def pymc3_dist(self, name, hypers):
        mu = self.mu
        std = self.std
        if(len(hypers) == 1):
                hyper_dist = hypers[0][0]
                hyper_name = hypers[0][1]
                idx = hypers[0][2]
                if(idx == 0):
                    mu = hyper_dist.pymc3_dist(hyper_name, [])
                else:
                    std = hyper_dist.pymc3_dist(hyper_name, [])
        elif(len(hypers) == 2):
                hyper_dist_1 = hypers[0][0]
                hyper_name_1 = hypers[0][1]
                hyper_dist_2 = hypers[1][0]
                hyper_name_2 = hypers[1][1]
                mu = hyper_dist_1.pymc3_dist(hyper_name_1, [])
                std = hyper_dist_2.pymc3_dist(hyper_name_2, [])

                
        if(self.num_elements==-1):
            return pm.Normal(name, mu=mu, sigma=std)
        else:
            return pm.Normal(name, mu=mu, sigma=std, shape=self.num_elements)
    def get_params(self):
    	return [self.mu, self.std]

    def scipy_dist(self, name):
        dist = (lambda siz : st.norm(self.mu, self.std).rvs(siz)) if self.num_elements == -1 else (lambda siz: st.norm(self.mu, self.std).rvs((self.num_elements, siz)))
        return name,dist


class Exponential(Continuous):
        
    """
    Class for the Exponential distribution 
    
    Attributes 
    -----------
    name: String of the name of the random variable
    
    lam: value for the lambda parameter. Default: 1
    
    num_elements: int specifying number of RV's
    
    is_hyper_param: Boolean specifying if this RV is used as a hyper parameter. Default: False
    
    """
    
    def __init__(self,name, lam=1, num_elements=-1, is_hyper_param=False):

        self.lam=lam
        self.name = name
        self.num_elements=num_elements
        self.is_hyper_param = is_hyper_param

    def pymc3_dist(self, name, hypers):
        lam = self.lam
        if(len(hypers) == 1):
            hyper_dist = hypers[0][0]
            hyper_name = hypers[0][1]
            lam = hyper_dist.pymc3_dist(hyper_name, [])
        if(self.num_elements==-1):
            return pm.Exponential(name, lam=lam)
        else:
            return pm.Exponential(name, lam=lam, shape=self.num_elements)

    def get_params(self):
    	return [self.lam]
    	
    def scipy_dist(self, name):
        dist = (lambda siz : st.expon(self.lam).rvs(siz)) if self.num_elements == -1 else (lambda siz: st.expon(self.lam).rvs((self.num_elements, siz)))
        return name,dist


class Beta(Continuous):
    """
    Class for the Beta distribution 
    
    Attributes 
    -----------
    
    name: String of the name of the random variable
    
    alpha: value for the alpha parameter. Default: 1
    
    beta: value for the beta parameter. Default: 1
    
    num_elements: int specifying number of RV's
    
    is_hyper_param: Boolean specifying if this RV is used as a hyper parameter. Default: False
    
    """
    
    def __init__(self, name, alpha=1, beta=1, num_elements=-1, is_hyper_param=False):


        self.alpha = alpha
        self.name  = name
        self.beta  = beta
        self.num_elements=num_elements
        self.is_hyper_param = is_hyper_param

    def pymc3_dist(self, name, hypers):
        alpha = self.alpha
        beta  = self.beta
        if(len(hypers) == 1):
                hyper_dist = hypers[0][0]
                hyper_name = hypers[0][1]
                idx = hypers[0][2]
                if(idx == 0):
                    alpha = hyper_dist.pymc3_dist(hyper_name, [])
                else:
                    beta = hyper_dist.pymc3_dist(hyper_name, [])
        elif(len(hypers) == 2):
                hyper_dist_1 = hypers[0][0]
                hyper_name_1 = hypers[0][1]
                hyper_dist_2 = hypers[1][0]
                hyper_name_2 = hypers[1][1]
                alpha = hyper_dist_1.pymc3_dist(hyper_name_1, [])
                beta = hyper_dist_2.pymc3_dist(hyper_name_2, [])


        if(self.num_elements==-1):
            return pm.Beta(name, alpha=self.alpha, beta=self.beta)
        else:
            return pm.Beta(name, alpha=self.alpha, beta=self.beta, shape=self.num_elements)
	
    def get_params(self):
    	return [self.alpha, self.beta]
    	
    def scipy_dist(self, name):
        dist = (lambda siz : st.beta(self.alpha,self.beta).rvs(siz)) if self.num_elements == -1 else (lambda siz: st.beta(self.alpha,self.beta).rvs((self.num_elements, siz)))
        return name,dist




from privugger.transformer.PyMC3.type_decoration import *
from privugger.distributions.continuous import Continuous
from privugger.distributions.discrete import Discrete, Constant
from privugger.transformer.PyMC3.theano_types import TheanoToken
from privugger.transformer.PyMC3.program_output import *

import astor
import pymc3 as pm
import theano.tensor as tt
import arviz as az
import os
import importlib

## Create a global pymc3 model and list of priors
global_model  = None
global_priors = []

## Set global variables for indicating if there was a concat/stack before infer
concatenated     = False
stacked          = False
global_model_set = False

def _from_distributions_to_theano(input_specs, output):
    
    itypes = []
    otype = []
    
    
    if(input_specs == None):
        itypes.append(TheanoToken.float_matrix)
    else:
        for s in input_specs:
            if(isinstance(s, str)):
                if(s == "continuous"):
                    itypes.append(TheanoToken.float_vector)
                else:
                    itypes.append(TheanoToken.int_vector)

            elif(s.is_hyper_param):
                continue

            elif(issubclass(s.__class__, Continuous)):
                if(s.num_elements == -1):
                    itypes.append(TheanoToken.float_scalar)
                elif(s.num_elements==1):
                    itypes.append(TheanoToken.single_element_float_vector)
                else:
                    itypes.append(TheanoToken.float_vector)

            else:
                if(s.num_elements == -1):
                    itypes.append(TheanoToken.int_scalar)
                elif(s.num_elements==1):
                    itypes.append(TheanoToken.single_element_int_vector)
                else:
                    itypes.append(TheanoToken.int_vector)

    #NOTE: This gets the output type.
    if(type(output).__name__ ==  "type"):
        if(output.__name__ == "Float"):
           otype.append(TheanoToken.float_scalar)
        elif(output.__name__ == "Int"):
           otype.append(TheanoToken.int_scalar)
    elif(type(output).__name__ == "List"):
        if(output.output.__name__ == "Int"):  
           otype.append(TheanoToken.int_vector)
        elif(output.output.__name__ == "Float"): 
           otype.append(TheanoToken.float_vector)
    else:
        if(output.output.__name__ == "Int"):  
           otype.append(TheanoToken.int_matrix)
        elif(output.output.__name__ == "Float"): 
           otype.append(TheanoToken.float_matrix)

            

    return (itypes, otype)

def concatenate(distribution_a, distribution_b,  type_of_dist, axis=0):

    """
    
    Parameters
    ----------- 

    distribution_a: The first distribution
    
    distribution_b: The second distribution
    
    type_of_dist: String that specifies if it is a continuous or discrete distribution
    
    axis: Int value giving the axis to stack
     
    Returns
    -----------
    Type of distribution: String
    """
    
    #NOTE we just return a tuple and then actually concat later. First element is the distributions and second specify the axis and
    #if we are concatenating or stacking
    global global_model
    global global_model_set
    global global_priors
    
    if(not global_model_set):
        global_model     = pm.Model()
        global_priors    = []
        global_model_set = True
    with global_model as model:
        val = pm.math.concatenate( (distribution_a.pymc3_dist(distribution_a.name, []), distribution_b.pymc3_dist(distribution_b.name, [])), axis=axis )
        global_priors.append(val)

    global concatenated
    concatenated = True
    return type_of_dist
    #return ((distribution_a, distribution_b), (axis, "concat"))


def stack(distributions,  type_of_dist, axis=0):
    """
    
    Parameters
    -----------

    distributions: A list of distributions
     
    type_of_dist: String that specifies if it is a continuous or discrete distribution

    axis: Int value giving the axis to stack
     
    Returns
    -----------
    Type of distribution: String
    """
     
    #NOTE we just return a tuple and then actually stack later. First element is the distributions and second specify the axis and
    #if we are concatenating or stacking
    global global_model
    global global_model_set
    global global_priors
    
    if(not global_model_set):
        global_model     = pm.Model()
        global_priors    = []
        global_model_set = True
        
    with global_model as model:
        stacked_variables = []
        for i in range(len(distributions)):
            stacked_variables.append(distributions[i].pymc3_dist(distriutions[i].name, []))
            global_priors.append(pm.math.stack(stacked_variables, axis=axis))

    global stacked
    stacked = True
    return type_of_dist
    #return (distributions, (axis, "stack"))


def sample_prior(model, samples=50):

    """
    This method does the inference when provided a PyMC3 model

    Parameters
    -------------
    model : PyMC3 model
    
    samples : int number of samples

    Returns
    ------------
    
    Samples from the priors: Priors
    
    """
    with model as sample_prior:
        prior_checks = pm.sample_prior_predictive(samples=samples)

        return prior_checks
    
def infer(prog, cores=2 , chains=2, draws=500, method="pymc3", return_model=False):
    """
    
    Parameters
    -----------
    
    prog: the program type specified as a privugger.Program type
    
    cores: Int number of cores to use for sampling. Default 500
    
    chains: Int number of chains. Default 2

    draws: Int number of draws. Default 2

    method: String specifying which backend to use

    return_model: Boolean. Returns the probabilistic model if true and the trace if false

    Returns
    ----------
    Trace produced by the probabilistic programming inference: Arviz trace
    """
    data_spec      = prog.dataset
    output         = prog.output_type
    num_specs      = len(data_spec.input_specs)
    input_specs    = data_spec.input_specs
    program        = prog.program

    global global_priors
    global global_model

    global concatenated
    global stacked
    global global_model_set
    
    if not (concatenated or stacked or global_model_set):
        
        global_model = pm.Model()
        global_priors = []
        global_model_set = True

    # global_model = pm.Model()
    # global_priors = []
    
    #### ##################
    ###### Lift program ###
    #######################
    if method == "pymc3":
        if(program is not  None):
            ftp = FunctionTypeDecorator()
            decorators = _from_distributions_to_theano(input_specs, output)
            
            lifted_program = ftp.lift(program, decorators)
            lifted_program_w_import = ftp.wrap_with_theano_import(lifted_program)
        
            #print(astor.to_source(lifted_program_w_import))
        
            f = open("typed.py", "w")
            f.write(astor.to_source(lifted_program_w_import))
            f.close()
            import typed as t 
            importlib.reload(t)
            
            #################
            ## Create model #
            #################
            trace = None
            with global_model:
                
                priors = []
                hyper_params = []
                
                for idx in range(num_specs):
                    prior = input_specs[idx]

                    #This is for the case when our prior comes from a concatenated/stacked distribution
                    if(isinstance(prior, str)):
                        continue
                  
                    if(prior.is_hyper_param):
                        hyper_params.append((prior, prior.name))
                    else:
                        params = prior.get_params()
                        hypers_for_prior = []
                        for p_idx in range(len(params)):
                            p = params[p_idx]
                            if(isinstance(p, Continuous) or isinstance(p, Discrete)):
                                for hyper in hyper_params:
                                    if(p.name == hyper[1]):
                                        hypers_for_prior.append((hyper[0],hyper[1], p_idx))
                            
                        #priors.append(prior.pymc3_dist(prior.name, hypers_for_prior))
                        global_priors.append(prior.pymc3_dist(prior.name, hypers_for_prior))
                
                if(program is not None):
                    #output = pm.Deterministic("output", t.method(*priors) )
                    output = pm.Deterministic(prog.name, t.method(*global_priors) )

                # Add observations
                prog.execute_observations(prior, output)

                if(return_model):
                    return global_model
                else:
                    trace = pm.sample(draws=draws, chains=chains, cores=cores,return_inferencedata=True)

                concatenated     = False
                stacked          = False
                global_model_set = False
                del global_model
                del global_priors
                #global_model = pm.Model()
                return trace
            
    elif method == "scipy":
        if isinstance(program, str):
            import re
            f = open(program, "r")
            new = open("typed.py", "w")
            for l in f.readlines():
                res = re.findall("def [a-zA-Z]+\(", l)
                if len(res):
                    new.write(re.sub("def [a-zA-Z]+\(", "def method(", l))
                else:
                    new.write(l)
            f.close()
            new.close()
            import typed as t
            importlib.reload(t)
            f = t.method
        else:
            f = program
        priors = []
        trace = {}
        for idx in range(num_specs):
            name, dist = input_specs[idx].scipy_dist(input_sepcs[idx].name)
            dist = dist(draws)
            priors.append(dist)
            trace[name] = dist
        outputs = []
        for pi in list(zip(*priors)):
            if len(pi) == 1:
                pi = pi[0]
            outputs.append(f(*pi))
        trace["output"] = outputs
        return az.convert_to_inference_data(trace)
    else:
        raise TypeError("Unsupported probabilistic framework")


    



"""
Module describing possible distributions
"""
import numpy as np

POSSIBLE_INTS = [0,1,2,3,4]
POSSIBLE_FLOATS = [i for i in range(6,13)]

#Ints Value
BINOMIAL, BERNOULLI, GEOMETRIC, BETA_BINOMIAL, POISSON, DISCRETE_UNIFORM = range(6)

#Floats Value
NORMAL, UNIFORM, TRUNCATED_NORMAL = range(6,9)
BETA, EXPONENTIAL, LAPLACE, STUDENT_T = range(9, 13)
CAUCHY, GAMMA = range(13, 15)

#MINIMUM COVERAGE
MINIMUM_PERCANTAGE_COVERAGE = 5.0
MINIMUM_COVERAGE = lambda low, high: (MINIMUM_PERCANTAGE_COVERAGE/(high-low))

#Distributions_Support
class Support:
    BINOMIAL = (0, np.inf)
    BERNOULLI = (0,1)
    GEOMETRIC = (1, np.inf)
    BETA_BINOMIAL = (0,np.inf)
    POISSON = (0,np.inf)
    DISCRETE_UNIFORM = (-np.inf, np.inf)

    NORMAL = (-np.inf, np.inf)
    UNIFORM = (-np.inf, np.inf)
    TRUNCATED_NORMAL = (-np.inf, np.inf)
    BETA = (0,1)
    EXPONENTIAL = (0, np.inf) # Figure if this makes sense. Since if the range is (10, 30) than exponential has to either be shifted or it is useless
    LAPLACE = (-np.inf, np.inf)
    STUDENT_T = (-np.inf, np.inf)
    CAUCHY = (-np.inf, np.inf)
    GAMMA = (0, np.inf)